diff --git a/sys/platform/pc32/i386/pmap.c b/sys/platform/pc32/i386/pmap.c
index b07f2cf..3caebc2 100644
--- a/sys/platform/pc32/i386/pmap.c
+++ b/sys/platform/pc32/i386/pmap.c
@@ -980,6 +980,8 @@ pmap_unwire_pte_hold(pmap_t pmap, vm_page_t m, pmap_inval_info_t info)
 /*
  * After removing a page table entry, this routine is used to
  * conditionally free the page, and manage the hold/wire counts.
+ *
+ * WARNING:  This function can block
  */
 static int
 pmap_unuse_pt(pmap_t pmap, vm_offset_t va, vm_page_t mpte,
@@ -1556,6 +1558,7 @@ pmap_remove_entry(struct pmap *pmap, vm_page_t m,
 				break;
 		}
 	}
+	KKASSERT(pv);
 
 	rtval = 0;
 	if (pv) {
@@ -1589,17 +1592,21 @@ pmap_insert_entry(pmap_t pmap, vm_offset_t va, vm_page_t mpte, vm_page_t m)
 
 	TAILQ_INSERT_TAIL(&pmap->pm_pvlist, pv, pv_plist);
 	TAILQ_INSERT_TAIL(&m->md.pv_list, pv, pv_list);
+	++pmap->pm_generation;
 	m->md.pv_list_count++;
 
 	crit_exit();
 }
 
 /*
- * pmap_remove_pte: do the things to unmap a page in a process
+ * pmap_remove_pte: do the things to unmap a page in a process.
+ *
+ * WARNING: This function may block (via pmap_remove_entry/pmap_unuse_pt),
+ *	    callers using temporary pmaps must reload them.
  */
 static int
 pmap_remove_pte(struct pmap *pmap, unsigned *ptq, vm_offset_t va,
-	pmap_inval_info_t info)
+		pmap_inval_info_t info)
 {
 	unsigned oldpte;
 	vm_page_t m;
@@ -1644,10 +1651,10 @@ pmap_remove_pte(struct pmap *pmap, unsigned *ptq, vm_offset_t va,
 /*
  * pmap_remove_page:
  *
- *	Remove a single page from a process address space.
+ * Remove a single page from a process address space.
  *
- *	This function may not be called from an interrupt if the pmap is
- *	not kernel_pmap.
+ * This function may not be called from an interrupt if the pmap is
+ * not kernel_pmap.
  */
 static void
 pmap_remove_page(struct pmap *pmap, vm_offset_t va, pmap_inval_info_t info)
@@ -1662,6 +1669,7 @@ pmap_remove_page(struct pmap *pmap, vm_offset_t va, pmap_inval_info_t info)
 		ptq = get_ptbase(pmap) + i386_btop(va);
 		if (*ptq) {
 			pmap_remove_pte(pmap, ptq, va, info);
+			/* ptq invalid */
 		}
 	}
 }
@@ -1748,7 +1756,8 @@ pmap_remove(struct pmap *pmap, vm_offset_t sva, vm_offset_t eva)
 		}
 
 		/*
-		 * NOTE: pmap_remove_pte() can block.
+		 * NOTE: pmap_remove_pte() can block and wipe the temporary
+		 *	 ptbase.
 		 */
 		for (; sindex != pdnxt; sindex++) {
 			vm_offset_t va;
@@ -2049,12 +2058,28 @@ pmap_enter(pmap_t pmap, vm_offset_t va, vm_page_t m, vm_prot_t prot,
 	/*
 	 * Mapping has changed, invalidate old range and fall through to
 	 * handle validating new mapping.
+	 *
+	 * Since we have a ref on the page directory page pmap_pte()
+	 * will always return non-NULL.
+	 *
+	 * NOTE: pmap_remove_pte() can block and cause the temporary ptbase
+	 *	 to get wiped.  reload the ptbase.  I'm not sure if it is
+	 *	 also possible to race another pmap_enter() but check for
+	 *	 that case too.
 	 */
-	if (opa) {
+	while (opa) {
 		int err;
+
 		err = pmap_remove_pte(pmap, pte, va, &info);
 		if (err)
 			panic("pmap_enter: pte vanished, va: %p", (void *)va);
+		pte = pmap_pte(pmap, va);
+		origpte = *(vm_offset_t *)pte;
+		opa = origpte & PG_FRAME;
+		if (opa) {
+			kprintf("pmap_enter: Warning, raced pmap %p va %p\n",
+				pmap, (void *)va);
+		}
 	}
 
 	/*
@@ -2842,6 +2867,7 @@ pmap_remove_pages(pmap_t pmap, vm_offset_t sva, vm_offset_t eva)
 			pte = (unsigned *)vtopte(pv->pv_va);
 		else
 			pte = pmap_pte_quick(pmap, pv->pv_va);
+		KKASSERT(*pte);
 		if (pmap->pm_active)
 			pmap_inval_add(&info, pmap, pv->pv_va);
 
@@ -2889,7 +2915,7 @@ pmap_remove_pages(pmap_t pmap, vm_offset_t sva, vm_offset_t eva)
 		 */
 		if (save_generation != pmap->pm_generation) {
 			kprintf("Warning: pmap_remove_pages race-A avoided\n");
-			pv = TAILQ_FIRST(&pmap->pm_pvlist);
+			npv = TAILQ_FIRST(&pmap->pm_pvlist);
 		}
 	}
 	pmap_inval_flush(&info);
